{ generated from Template DLLIST__.PTT with :
TDLL_Data_Templ=Pointer
Empty_Data=nil
}
{
  Doubly linked lists.


  Copyright (C) 1998     Robert Zierer <rz1@stud.mw.tu-muenchen.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

$Revision: 1.0 $
$Date: 1999/05/02 07:56:36 $

}

unit Dllist__Pointer;

{ Bugs : Possible overflow of TDL_Pointer_List.FSize will not be checked !}
{$X+}

(* Previous version from  Jeff Atwood, JAtwood159@AOL.COM.*)
(* 4-22-98 RZ *)

interface
uses SysUtils;
type
TDLL_Pointer_Data = Pointer; {Could also point to something different}
 {user needn't know this}
DLL_Pointer_ElementPtr = ^TDLL_Pointer_Element;
TDLL_Pointer_Element = record
            next, prev : DLL_Pointer_ElementPtr;
            DLL_Data : TDLL_Pointer_Data;
            end; {}
TComparision = ( DLLstop, DLLunknown, DLLequal, DLLless, DLLgreater );
TSortComparision = DLLless .. DLLgreater ;
TSortOrder = ( DLLrising, DLLnone, DLLfalling );
TDLLRecurseFunctionResult = ( DLLcontinue, DLLstopExec, DLLnotImplemented );
TDLLRecurseFunction = function( item: TDLL_Pointer_Data; loc: longint ): TDLLRecurseFunctionResult;
EDLLDelete = class(Exception);
TDL_Pointer_List = class(TObject)
   private
      Top, Bottom, Current : DLL_Pointer_ElementPtr;
      FSize : Longint;
      FoundLoc : longint;
      FindItem : TDLL_Pointer_Data;
      FoundPlace : DLL_Pointer_ElementPtr;
      FSortOrder : TSortOrder;
      procedure SetSortOrder( SO : TSortOrder);
   protected
      function SortRisingFalling(CompareState:TSortComparision): Boolean; virtual;{can be made better}
      function Sort: Boolean; virtual;
   public
      WarnOnDel : Boolean; {if true, checks at del, if DLL_Data empty; if not --> Exception;
                            default : false !!}
      property SortOrder : TSortOrder read FSortOrder write SetSortOrder;
               {if SortOrder changes, --> resorting necessary}
      property Size : longint read FSize;
      constructor create;
      destructor destroy; override;
      function IsEmpty: Boolean;
      function SizeOf : longint; virtual;
      function InsertBottom( item: TDLL_Pointer_Data ): Boolean;
      function InsertTop( item: TDLL_Pointer_Data ): Boolean;
      function InsertCurrent( item: TDLL_Pointer_Data ): Boolean;
      function Insert( item: TDLL_Pointer_Data ): Boolean;
      function FindFirst( item: TDLL_Pointer_Data; var absLoc: longint): Boolean;
      function FindNext( var absLoc: longint): Boolean;
      function Delete: Boolean;
      function MoveFirst: Boolean;
      function MoveLast: Boolean;
      function MoveNext: Boolean;
      function MovePrevious: Boolean;
      function Seek( absLoc: longint ): Boolean;
      function GetCurrentDataPointer( var item: TDLL_Pointer_Data ): Boolean;
      function SetCurrentDataPointer( item: TDLL_Pointer_Data ): Boolean;
      function Compare( a, b : TDLL_Pointer_Data ) : TComparision; virtual;
      procedure Pack;
      function Recurse( r: TDLLRecurseFunction;var loc: longint; endloc : longint ): boolean;
   end;


implementation

{ set up the TDL_Pointer_List object with default values }
constructor TDL_Pointer_List.create;
begin
  inherited create;
  top := nil;
  bottom := nil;
  current := nil;
  FSize := 0;
  FSortOrder := DLLnone;
  WarnOnDel := false;
  FoundLoc := 0;
  FindItem := nil;
  FoundPlace := Nil;
end;

{ destroy the entire list, cell by cell }
destructor TDL_Pointer_List.destroy;
var
  curCell,
  nextCell: DLL_Pointer_ElementPtr;
begin
  if WarnOnDel=true then begin
    {check, if list to del}
    curCell := Top;
    while not (curCell = nil) do begin
      if ( curCell^.DLL_Data <> Nil ) then
        raise EDLLDelete.Create('Element, that is to delete contains data !');
      curCell := curCell^.next;
    end;
  end;
  {delete each element}
  curCell := top;
  while not (curCell = nil) do begin
    nextCell := curCell^.next;
    dispose(curCell);
    curCell := nextCell;
  end;
  top := nil;
  bottom := nil;
  current := nil;
  inherited destroy;
end;

{ returns true if the list has no cells }
function TDL_Pointer_List.isEmpty: Boolean;
begin
  result := (FSize = 0);
end;

{ returns number of bytes used for list }
function TDL_Pointer_List.SizeOf : longint;
begin
   result := FSize * System.SizeOf( TDLL_Pointer_Element) {von Liste};
end;

{ insert cell at bottom of list }
function TDL_Pointer_List.InsertBottom(item: TDLL_Pointer_Data): Boolean{true, if possible};
var
  newCell: DLL_Pointer_ElementPtr;
begin
  result := true;
  New(newCell);
  newCell^.DLL_Data := item;
  newCell^.prev := bottom;
  newCell^.next := nil;
  { special case: this is first cell added }
  if bottom = nil then
    top := newCell
  else
    bottom^.next := newCell;
  bottom := newCell;
  inc(FSize);
  FSortOrder := DLLnone;
end;

{ insert cell at top of list }
function TDL_Pointer_List.InsertTop(item: TDLL_Pointer_Data): Boolean{true, if possible};
var
  newCell: DLL_Pointer_ElementPtr;
begin
  result := true;
  New(newCell);
  newCell^.DLL_Data := item;
  newCell^.prev := nil;
  newCell^.next := top;
  { special case: this is first cell added }
  if top = nil then
    bottom := newCell
  else
    top^.prev := newCell;
  top := newCell;
  inc(FSize);
  FSortOrder := DLLnone;
end;

{ insert cell after current item }
function TDL_Pointer_List.InsertCurrent(item: TDLL_Pointer_Data): Boolean {true,if current exists};
var
  newCell: DLL_Pointer_ElementPtr;
begin
  if (current = nil) then
    result := False
  else begin
    New(newCell);
    newCell^.DLL_Data := item;
    newCell^.prev := current;
    newCell^.next := current^.next;
    { special case: current cell is last cell }
    if current^.next = nil then
      bottom := newCell
    else
      current^.next^.prev := newCell;
    current^.next := newCell;
    inc(FSize);
    result := True;
    FSortOrder := DLLnone;
  end;
end;

{ Look for item in DLL_Data field. Starts at top of list }
{ and looks at every item until a match is found.    }
{ if found, makes matched cell current, and returns  }
{ absolute location of match where 1 = top.          }
function TDL_Pointer_List.FindFirst(item: TDLL_Pointer_Data; var absLoc: longint): Boolean{true, if found};
var
  curCell: DLL_Pointer_ElementPtr;
  cnt: longInt;
begin
  result := False;  FoundLoc := 0; FindItem := nil;
  curCell := top;
  cnt := 0;
  absLoc := 0;
  while not (curCell = nil) do begin
    inc(cnt);
    if Compare( curCell^.DLL_Data, item ) = DLLequal then begin
      absLoc := cnt;
      current := curCell;
      result := True;
      {for Findnext}
      FoundLoc := cnt;
      FindItem := item;
      FoundPlace := current;
      exit;
    end;
    curCell := curCell^.next;
  end;
end;

{continues search of FindFirst}
function TDL_Pointer_List.FindNext( var absLoc: longint): Boolean;
var
  curCell: DLL_Pointer_ElementPtr;
  cnt: longInt;
begin
   result := false;
   if (FoundLoc<>0) and (FindItem<>nil) and (FoundPlace<>Nil) then begin
      curCell := FoundPlace^.next;
      cnt := FoundLoc;
      absLoc := 0;
      while not (curCell = nil) do begin
         inc(cnt);
         if Compare( curCell^.DLL_Data, FindItem ) = DLLequal then begin
            absLoc := cnt;
            current := curCell;
            result := True;
            {for Findnext}
            FoundLoc := cnt;
            FoundPlace := current;
            exit;
          end;
          curCell := curCell^.next;
       end;
  end;
end;

{ delete the current cell }
{ ... and sets current to nil}
function TDL_Pointer_List.Delete: Boolean{true,if current exists};

  procedure goto_exitDelete;
  begin
    Delete := True;
    dispose( current );
    current := nil;
    dec( FSize );
    if FSize = 0 then begin
      top := nil;
      bottom := nil;
    end;
  end;

begin
  {check, if element to del}
  if (current^.DLL_Data <> nil) and (WarnOnDel=true) then
     raise EDLLDelete.Create('Element, that is to delete contains data !');
  { we can only delete the current record }
  if current = nil then
    Result := False
  else begin
    { see if list has one item }
    if Size = 1 then begin
      top := nil;
      bottom := nil;
    end else
    { see if we're at the top of list }
    if current^.prev = nil then begin
      top := current^.next;
      top^.prev := nil;
    end else
    { see if we're at the bottom of list }
    if current^.next = nil then begin
      bottom := current^.prev;
      bottom^.next := nil;
    end else begin
      { we must be in middle of list of size > 1 }
      current^.prev^.next := current^.next;
      current^.next^.prev := current^.prev;
    end;
    goto_exitDelete;
  end;
end;

{*******************************************************************************************}
{*******************************************************************************************}
{*******************************************************************************************}
{*******************************************************************************************}
{*******************************************************************************************}

{ make first value in list current }
function TDL_Pointer_List.MoveFirst: Boolean{true, if list not empty};
begin
  if top = nil then
    result := False
  else begin
    current := top;
    result := True;
  end;
end;

{ make last value in list current }
function TDL_Pointer_List.MoveLast: Boolean{true, if list is not empty};
begin
  if bottom = nil then
    result := False
  else begin
    current := bottom;
    result := True;
  end;
end;

{ make next value in list current }
function TDL_Pointer_List.MoveNext: Boolean{true, if next element exists and current exists};
begin
  if (current = nil) or (current^.next = nil) then
    result := False
  else begin
    current := current^.next;
    result := True;
  end
end;

{ make previous value in list current }
function TDL_Pointer_List.MovePrevious: Boolean{true, if previous element exists and current exists};
begin
  if (current = nil) or (current^.prev = nil) then
    result := False
  else begin
    current := current^.prev;
    result := True;
  end;
end;

{ return DLL_Data from current list position }
function TDL_Pointer_List.GetCurrentDataPointer(var item: TDLL_Pointer_Data): Boolean{true,if current exists};
begin
  if (current = nil) then
    result := False
  else begin
    item := current^.DLL_Data;
    result := True;
  end;
end;

{ sets DLL_Data in Current element}
function TDL_Pointer_List.SetCurrentDataPointer(item: TDLL_Pointer_Data): Boolean{true,if current exists};
begin
  if (current = nil) then
    result := False
  else begin
    current^.DLL_Data := item;
    result := True;
  end;
end;

{ make current the absolute cell N in the list }
{ where top = 1 }
function TDL_Pointer_List.Seek( absloc: longint ): Boolean{true, if successful};
var
  curCell: DLL_Pointer_ElementPtr;
  cnt: longint;
begin
  result := False;
  cnt := 0;
  if absloc <= 0 then
    exit;
  curCell := top;
  while not (curCell = nil) do begin
    inc(cnt);
    if cnt = absloc then begin
      current := curCell;
      result := True;
      exit;
    end;
    curCell := curCell^.next;
  end;
end;

{compares two elements} {to be overriden by it's descendant !!}
function TDL_Pointer_List.Compare( a, b : TDLL_Pointer_Data ) : TComparision;
begin
   {only a dummy for later Objects}
   result := DLLunknown;
end;

procedure TDL_Pointer_List.SetSortOrder( SO : TSortOrder);
begin
   if SO <> FSortOrder then begin
      {resort list, if SO <> previous state}
      FSortOrder := SO;
      if Sort = false then
        FSortOrder := DLLnone;
   end;
end;

{ inserts element in sorted or unsorted list}
function TDL_Pointer_List.Insert( item: TDLL_Pointer_Data ): Boolean{true, if possible};
var oldSortOrder : TSortOrder;
begin
   if FSortOrder = DLLnone then begin
      if current <> nil then
         result := InsertCurrent(item){insert after Current}
      else
         result := InsertBottom(item); {else somewhere}
   end else begin {sorted}
      {insert where it fits; the better at the beginning (Bubble-Sort)}
      oldSortOrder := SortOrder;
      result := InsertTop(item); {InsertTop changes SortOrder to none}
      if result = true then begin
         FSortOrder := oldSortOrder; {and needs with it resorting}
         result := Sort;{result is needed here, therefore no SortOrder:=oldSortOrder}
         if result = false then
            FSortOrder := DLLnone; {for security reasons}
      end;
   end;
end;

{sorts list}
function TDL_Pointer_List.Sort: Boolean;{true, if successfull}
begin
   Result := false;
   case FSortOrder of
         DLLfalling : begin
                   result := SortRisingFalling(DLLless);
                   end;
         DLLrising : begin
                  result := SortRisingFalling(DLLgreater);
                  end;
         DLLnone : result := true;
   end;
end;

{sorts list in rising order, if CompareState=DLLgreater;
 sorts list in falling order, if CompareState=DLLless}
function TDL_Pointer_List.SortRisingFalling(CompareState:TSortComparision): Boolean; {Bubble2 ??}
var curCell: DLL_Pointer_ElementPtr;
    obergrenze, marke,  index : longint;
    swaphelp : TDLL_Pointer_Data;
    comp : TComparision;
begin
   obergrenze := FSize;
   while obergrenze >1 do begin
      curCell := Top;
      marke :=1;
      for index :=1 to obergrenze-1 do begin
         comp := Compare(curCell^.DLL_Data , curCell^.Next^.DLL_Data);
         case comp of
            DLLstop,DLLunknown : begin
                           result := false;
                           exit;
                           end;
            DLLless,DLLgreater : if comp = CompareState then begin
                              swaphelp := curCell^.Next^.DLL_Data;
                              curCell^.Next^.DLL_Data := curCell^.DLL_Data;
                              curCell^.DLL_Data:=swaphelp;
                              marke:=index;
                           end;
         end;
         curCell := curCell^.Next;
      end;
      obergrenze := marke;
   end;
   result := true;
end;

{erases elements with DLL_Data = nil}
procedure TDL_Pointer_List.Pack;
var curCell,
    CurrentSave : DLL_Pointer_ElementPtr;
begin
   curCell := top;
   while (curCell <> nil) do begin
      if curCell^.DLL_Data = nil then begin
         if Current <> CurCell then begin
            CurrentSave := Current; {save Current temporarily}
            Current := curCell;
         end else
            CurrentSave := nil; {Current will be del by Delete}
         curCell := curCell^.Prev;{otherwise curCell would point to void}
         Delete;
         Current := CurrentSave;
      end;
      curCell := curCell^.Next;
   end;
end;

{recurses through list,
 applying r to all elements till list ends or r results DLLstopExec (or notImplemented)}
function TDL_Pointer_List.Recurse( r: TDLLRecurseFunction;var loc: longint; endloc : longint ): boolean;
                               {true,if all done; false, loc shows number of Element not done}
              {loc is first Element, 0, if from beginning;
               endloc is last Element, 0, if all till listend }
var curCell: DLL_Pointer_ElementPtr;
    tempres: TDLLRecurseFunctionResult;
begin
   case loc of
      0 : begin {search all cells}
          result := true;
          curCell := top;
          loc := 1;
          end;
      else    {search startcell}
          result := Seek(loc);
          if result = true then
             curCell := Current;
   end;
   if result = true then begin
      tempres := DLLcontinue;
      while (curCell <> nil) and (tempres = DLLcontinue) and ((loc<=endloc)or(loc=0)) do begin
         tempres := r( curCell^.DLL_Data, loc );
         inc(loc);
      end;
      if tempres = DLLContinue then
         result := true;
   end;
end;

initialization
end.
