{
  Dynamic array class with elements up to 64K and constant elementsize.


  Copyright (C) 1998     Robert Zierer <rz1@stud.mw.tu-muenchen.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

$Revision: 1.0 $
$Date: 1999/05/02 07:56:36 $

}

unit Dynarray;

(* 4-22-98 RZ *)

{$define InitializeWithZeros} {preinitialize any Block with Zero's}

interface
uses DLList__Pointer, SysUtils, Winprocs;
type
  TDynLimArrayElementPtr = Pointer;{or any other type}
  EDynLimArray = class(Exception);
  TDynLimArray = class(TDL_Pointer_List)
     private
        ElementSizeSet : boolean;
        FDimensionated : boolean;
        FElementSize: word;
        Elements_per_block: word;{number of elements in the blocks}
        Blocks : longint; {number of blocks with Elements_per_block elements}
        LastBlockContents: word;{number of elements in the last block}
        FLowIndex : longint;
        FHighIndex : longint;
        procedure SetElementSize( s : word );
        procedure SetLowIndex( s: longint );
        procedure SetHighIndex( s: longint );
        function GetIndexSize: longint;
     protected
     public
        property Dimensionated : boolean read FDimensionated;
        property ElementSize : word read FElementSize write SetElementSize;
        property LowIndex : longint read FLowIndex write SetLowIndex;
        property HighIndex : longint read FHighIndex write SetHighIndex;
        property IndexSize : longint read GetIndexSize;
        constructor create;
        destructor destroy; override;
        function SizeOf : longint; override;
        procedure Dim( low, high: longint );
        procedure ReDim( low, high: longint );
        function GetElementPtr( item: longint ): TDynLimArrayElementPtr;
        procedure Clear;
     end;

const MaxElementSize = 65519;

implementation

{ set up the TDynLimArray object with default values }
constructor TDynLimArray.create;
begin
   inherited create;
   {of TDLList__Pointer}
   WarnOnDel := false;
   SortOrder := DLLnone;
   {own}
   ElementSizeSet := false;
   FElementSize := 0;
   Elements_per_block :=0;
   FDimensionated := false;
   FLowIndex := 0;
   FHighIndex := 0;
   Blocks :=0;
   LastBlockContents := 0;
end;

{ destroys the entire TDynLimArray, cell by cell }
destructor TDynLimArray.destroy;
begin
  Clear;
  inherited destroy;
end;

{ returns number of bytes used for array }
function TDynLimArray.SizeOf : longint;
begin     {from Array}
   result := inherited SizeOf + ( Blocks*Elements_per_Block+LastBlockContents)*FElementSize;
end;

{delivers number of elements in Array}
function TDynLimArray.GetIndexSize: longint;
begin
   if not FDimensionated then result := 0
   else result := HighIndex - LowIndex + 1;
end;

{Erases the whole Array}
procedure TDynLimArray.Clear;
var i : longint;
    p : TDynLimArrayElementPtr;
begin
   if blocks>0 then
     for i := 1 to Blocks do begin
       MoveFirst;
       GetCurrentDataPointer( p );
       freemem( p, Elements_per_block*FElementSize);
       SetCurrentDataPointer( nil );
       Delete;
     end;
   if LastBlockContents > 0 then begin
     MoveFirst;
     GetCurrentDataPointer(p);
     freemem( p,LastBlockContents*FElementSize);
     SetCurrentDataPointer( nil);
     Delete;
   end;
   FDimensionated := false;
   FLowIndex := 0;
   FHighIndex := 0;
   Blocks :=0;
   LastBlockContents := 0;
end;

{sets ElementSize only once}
procedure TDynLimArray.SetElementSize( s : word );
begin
   if ElementSizeSet = true then
      raise EDynLimArray.Create('Can set ElementSize only once !')
   else begin
      if s<1 then raise EDynLimArray.Create('Wrong ElementSize');
      if s>MaxElementSize then raise EDynLimArray.Create('ElementSize to big!');
      FElementSize := s;
      ElementSizeSet := true;
      {calc number of elements in a block}
      Elements_per_block :=trunc( MaxElementSize / FElementSize );
   end;
end;

{sets low ArrayIndex and restructures the Array}
procedure TDynLimArray.SetLowIndex( s: longint );
var oldnewLastBlock, old, q : TDLL_Pointer_Data;
    p : TDynLimArrayElementPtr;
    newBlocks, newLastBlockContents, i, fullBlocks, Offset, oldBlock : longint;
    ok, valid : boolean;

     procedure InsertBlock;
     begin
       if oldBlock = 1 then
         InsertTop( p )
       else begin
         Seek(oldBlock-1);
         InsertCurrent(p);
       end;
     end;


    procedure ProcessNext;
    var finished : boolean;
    begin
      finished := false;
      repeat
        {right in the middle}
        {copy 2nd halfblock}
        if i+Elements_per_block-1 >= FHighIndex then begin {at the end}
          getmem(p, (FHighIndex-i+1)*FElementSize);
          q := ptr( seg(old), ofs(old) + (Elements_per_block-Offset)*FElementSize );
          if i+Elements_per_block-Offset-1 >= FHighIndex then begin
            {only 2nd halfblock incomplete}
            hmemcpy( p, q, (FHighIndex-i+1)*FElementSize );
            InsertBlock;
            {dealloc old block}
            MoveNext;
            Freemem( old, LastBlockContents*FElementSize );
            SetCurrentDataPointer( nil );
            Delete;
          end else begin
            {2nd halfblock}
            hmemcpy( p, q, (Elements_per_block-Offset)*FElementSize );
            InsertBlock;
            {dealloc old block}
            MoveNext;
            Freemem( old, Elements_per_Block*FElementSize );
            SetCurrentDataPointer( nil );
            Delete;
            {1st halfblock incomplete}
            inc( oldBlock );
            Seek( oldBlock );
            GetCurrentDataPointer( old );
            q := ptr( seg(p), ofs(p) + Offset*FElementSize );
            hmemcpy( p, old, LastBlockContents*FElementSize );
            {dealloc old last block}
            Freemem( old, LastBlockContents*FElementSize );
            SetCurrentDataPointer( nil );
            Delete;
          end;
          finished := true;
        end else begin
          {normal block in the middle}
          getmem( p, Elements_per_block*FElementSize );
          hmemcpy( p, q, (Elements_per_block-Offset)*FElementSize );
          InsertBlock;
          {dealloc old block}
          MoveNext;
          Freemem( old, Elements_per_block*FElementSize );
          SetCurrentDataPointer( nil );
          Delete;
          {1st halfblock}
          inc( oldBlock );
          Seek( oldBlock );
          GetCurrentDataPointer( old );
          q := ptr( seg(p), ofs(p) + Offset*FElementSize );
          hmemcpy( q, old, Offset*FElementSize );
        end;
      until finished;
    end;

begin
  if s = FLowIndex then exit;
  if not FDimensionated then raise EDynLimArray.create
                                   ('Couldn''t redimensionate Array before dimensionating !');
  if s > FHighIndex then raise EDynLimArray.create('LowIndex > HighIndex !');
  newBlocks := trunc((FHighIndex-s+1)/Elements_per_block);
  newLastBlockContents := FHighIndex-s+1-newBlocks*Elements_per_block;
  if s < FLowIndex then begin {delete}
    fullBlocks := trunc( (s-FLowIndex)/Elements_per_block );
    Offset := Elements_per_block-(FLowIndex-s-fullBlocks*Elements_per_block);
    if newLastBlockContents = LastBlockContents then begin
      {only delete full blocks at the beginning}
      for i := 1 to fullBlocks do begin
        MoveFirst;
        GetCurrentDataPointer( p );
        freemem( p, Elements_per_block*FElementSize);
        SetCurrentDataPointer( nil );
        Delete;
      end;
    end else begin
      {delete normal}
      if fullBlocks >0 then
        for i := 1 to fullBlocks do begin
          MoveFirst;
          GetCurrentDataPointer( p );
          freemem( p, Elements_per_block*FElementSize);
          SetCurrentDataPointer( nil );
          Delete;
        end;
      oldBlock := 1;
      MoveFirst;
      GetCurrentDataPointer(old);
      i:=s;
      ProcessNext;
    end;
  end else begin {append}
    fullBlocks := trunc( (FLowIndex-s)/Elements_per_block );
    Offset := FLowIndex-s-fullBlocks*Elements_per_block;
    if newLastBlockContents = LastBlockContents then begin
      {only append full blocks at the beginning}
      for i := 1 to fullBlocks do begin
        getmem( p, Elements_per_block*FElementSize);
        InsertTop( p );
      end;
    end else begin
      {append normal}
      if fullBlocks > 0 then
        for i := 1 to fullBlocks do begin
          getmem( p, Elements_per_block*FElementSize);
          InsertTop( p );
        end;
      oldBlock := fullBlocks+1;
      Seek(oldBlock);
      GetCurrentDataPointer(old);
      i:=fullBlocks*Elements_per_block+s;
      {first halfblock of 1st block}
      if i+Elements_per_block-1 >= FHighIndex then begin
        {Special case: only one short 1st halfblock}
        getmem(p,(FHighIndex-i+1)*FElementSize);
        InsertBlock;
        q := ptr( seg(p), ofs(p) + Offset*FElementSize );
        hmemcpy( q, old, (FHighIndex-i+1)*FElementSize );
        Seek(oldBlock+1);
        SetCurrentDataPointer( nil );
        Delete;
        freemem(old,LastBlockContents*FElementSize);
      end else begin
        {normal beginning} {normal 1st halfblock}
        getmem(p,Elements_per_block*FElementSize);
        InsertBlock;
        q := ptr( seg(p), ofs(p) + Offset*FElementSize );
        hmemcpy( q, old, Offset*FElementSize );
        ProcessNext;
      end;
    end;
  end;
  FLowIndex := s;
  LastBlockContents := newLastBlockContents;
  Blocks := newBlocks;
end;

{sets high ArrayIndex and restructures the Array}
procedure TDynLimArray.SetHighIndex( s: longint );
var oldnewLastBlock : TDLL_Pointer_Data;
    p, old : TDynLimArrayElementPtr;
    newBlocks, newLastBlockContents, i : longint;
    ok, valid : boolean;

    function Comparision(a,b:longint):TComparision;
    begin
      if a<b then result := DLLless
      else if a>b then result := DLLgreater
           else result := DLLequal;
    end;

begin
  if s = FHighIndex then exit;
  if not FDimensionated then raise EDynLimArray.create
                                   ('Couldn''t redimensionate Array before dimensionating !');
  if s < FLowIndex then raise EDynLimArray.create('HighIndex < LowIndex !');
  {calc new Index}
  newBlocks := trunc((s-FLowIndex+1)/Elements_per_block);
                           {new number of full blocks}
  newLastBlockContents := s-FLowIndex+1-newBlocks*Elements_per_block;
  {save blocks that will be needed}
  if s > FHighIndex then begin {inc Index}
    {save last old block and delete from list}
    valid := MoveLast;
    ok := GetCurrentDataPointer(oldnewLastBlock) and valid;
    {del lst old block from list}
    if (oldnewLastBlock<>nil) and ok then begin
      valid := SetCurrentDataPointer(nil);
      ok := Delete and valid;
      if not ok then {couldn't del last old element}
        raise EDynLimArray.create('Internal Error : Can''t delete last element');
    end else {couldn't positionate ??}
        if oldnewLastBlock = nil then {last element of list doesn't exist ???}
          raise EDynLimArray.create('Internal Error : Error in last element of list');
  end else begin {dec index}
    {save last new block, if needed}
    if newLastBlockContents <> 0 then begin
      valid := Seek(newBlocks+1);
      ok := GetCurrentDataPointer(oldnewLastBlock) and valid;
      if (not ok)or(oldnewLastBlock=nil) then {couldn't find last new element}
        raise EDynLimArray.create('Internal Error : Can''t find last element');
      {del last old block from list, if blocks equal}
      if newBlocks = Blocks then begin
        valid := SetCurrentDataPointer(nil);
        ok := Delete and valid;
        if not ok then {couldn't del last old element}
          raise EDynLimArray.create('Internal Error : Can''t delete last element');
      end;
    end;
  end;
  {detect type of Operation}
  case Comparision(newBlocks,Blocks) of
  DLLgreater: begin {newBlocks>Blocks}
           {at least one full block to append}
           for i := Blocks+1 to newBlocks do begin
             getmem( p, Elements_per_block*FElementSize );
             InsertBottom( p );
             if i = Blocks+1 then begin
               {transmit content of last block}
               hmemcpy( p, oldnewLastBlock, LastBlockContents*FElementSize );
               {dealloc last block}
               freemem( oldnewLastBlock, LastBlockContents*FElementSize);
             end;
           end;
           if newLastBlockContents <>0 then begin
             getmem( p, newLastBlockContents*FElementSize);
             InsertBottom( p );
           end;
           end;
  DLLequal: begin {newBlocks=Blocks; Changes only in last block; but not full }
         {differ cases : newLastBlockContents =0, <,>LastBlockContents}
         if newLastBlockContents <> 0 then begin
           getmem( p, newLastBlockContents*FElementSize);
           if newLastBlockContents>LastBlockContents then begin
             {newLastBlockContents in any case greater than 0 and block not full}
             {transmit content of last block}
             hmemcpy( p, oldnewLastBlock, LastBlockContents*FElementSize );
           end else begin
             {newLastBlockContents < LastBlockContents, but not 0}
             {transmit content of last block}
             hmemcpy( p, oldnewLastBlock, newLastBlockContents*FElementSize );
           end;
           {insert}
           InsertBottom( p );
           {del last block}
           freemem( oldnewLastBlock, LastBlockContents*FElementSize);
         end else begin {del only last block}
           valid := MoveLast;
           ok := GetCurrentDataPointer( p ) and valid;
           if ok then begin
             freemem( p, LastBlockContents*FElementSize );
             SetCurrentDataPointer( nil );
             if Delete=false then
               raise EDynLimArray.Create
                         ('Internal Error : Can''t delete last old element');
           end else
               raise EDynLimArray.Create
                            ('Internal Error : Can''t find last old element');
         end;
         end;
  DLLless: {newBlocks<Blocks}
        {at least one full block to del}
        for i := Blocks+1 downto newBlocks+1 do begin
          if i = Blocks+1 then begin
            {del last old block}
            if LastBlockContents <>0 then begin
              valid := MoveLast;
              ok := GetCurrentDataPointer( p ) and valid;
              if ok then begin
                freemem( p, LastBlockContents*FElementSize );
                SetCurrentDataPointer( nil );
                if Delete=false then
                  raise EDynLimArray.Create
                         ('Internal Error : Can''t delete last old element');
              end else
                raise EDynLimArray.Create
                            ('Internal Error : Can''t find last old element');
            end;
          end else if i = newBlocks+1 then begin
            {last new block resp. not existing block at the end}
            valid := MoveLast;
            ok := GetCurrentDataPointer( old ) and valid;
            if ok then begin
              if newLastBlockContents <>0 then begin
                getmem( p, newLastBlockContents*FElementSize);
                {transmit content of last block}
                hmemcpy( p, old, newLastBlockContents*FElementSize );
              end;
              {dealloc old block}
              freemem( old, Elements_per_block*FElementSize );
              SetCurrentDataPointer( nil );
              if Delete=false then
                raise EDynLimArray.Create('Internal Error : Can''t delete element');
              {and insert new one}
              if newLastBlockContents <>0 then InsertBottom( p );
            end else
              raise EDynLimArray.Create('Internal Error : Can''t find element');
          end else begin
            {del any block}
            valid := MoveLast;
            ok := GetCurrentDataPointer( p ) and valid;
            if ok then begin
              freemem( p, Elements_per_block*FElementSize );
              SetCurrentDataPointer( nil );
              if Delete=false then
                raise EDynLimArray.Create('Internal Error : Can''t delete element');
            end else
              raise EDynLimArray.Create('Internal Error : Can''t fine element');
          end;
        end;
  end;
  {correct all values}
  Blocks := newBlocks;
  LastBlockContents := newLastBlockContents;
  {adapt Index}
  FHighIndex := s;
end;

{delivers a pointer to the element}
function TDynLimArray.GetElementPtr( item: longint ): TDynLimArrayElementPtr;{nil on Error}
var p : TDynLimArrayElementPtr;
begin
  if not FDimensionated then raise EDynLimArray.create
                                 ('Can''t access Array before dimensionating !');
  if (item > FHighIndex) or (FElementSize=0) or (item < FLowIndex) then result := nil
   else begin
      seek( trunc((item-FLowIndex)/Elements_per_block)+1);
      if GetCurrentDataPointer( p ) then begin
         inc( PtrRec(p).ofs,(item-FLowIndex-trunc((item-FLowIndex)/Elements_per_block)*Elements_per_block)*FElementSize );
         result := p;
      end else
         result := nil;
   end;
end;

{ dimensionates the Array}
procedure TDynLimArray.Dim( low, high: longint );
var i : longint;
    p : TDynLimArrayElementPtr;
    result : boolean;
begin
   result := true;
   if not ElementSizeSet then raise EDynLimArray.create('ElementSize not set');
   if FDimensionated then raise EDynLimArray.create
                   ('Can''t dimensionate a dimensionated Array again !');
   if high<low then raise EDynLimArray.create('Indexing Error');
   Blocks := trunc((high-low+1)/Elements_per_block);
   if Blocks > 0 then
     for i := 1 to Blocks do begin
       getmem( p, Elements_per_block*FElementSize );
       {$ifdef InitializeWithZeros}
       FillChar( p^, Elements_per_block*FElementSize, #0 ); {fill with Zero's}
       {$endif}
       result := InsertBottom( p ) and result;
       if not result then {not enough mem avail for list}
         raise EOutOfMemory.Create('No memory for list');;
     end;
   LastBlockContents := high-low+1-Blocks*Elements_per_block;
   if LastBlockContents <>0 then begin
     getmem( p, LastBlockContents*FElementSize);
     {$ifdef InitializeWithZeros}
     FillChar( p^, LastBlockContents*FElementSize, #0 ); {fill with Zero's}
     {$endif}
     result := InsertBottom( p ) and result;
   end;
   if result then begin
     FDimensionated := true;
     FLowIndex := low;
     FHighIndex := high;
   end;
end;

{ redimensionates the Array}
procedure TDynLimArray.ReDim( low, high: longint );
begin
  if not FDimensionated then raise EDynLimArray.create
                                 ('Can''t access Array before dimensionating !');
  if low>high then raise EDynLimArray.create
                                 ('low Index > high Index');
  if (low>FHighIndex) or (high<FLowIndex) then begin
    Clear;
    Dim( low, high);
  end else begin
    SetLowIndex( low );
    SetHighIndex( high );
  end;
end;

initialization
end.

