{
  Miscellaneous things for the Yacc parser:

    - simple *.res-file routines.


  Copyright (c) 1990-92  Albert Graef <ag@muwiinfa.geschichte.uni-mainz.de>
  Copyright (C) 1998     Robert Zierer <rz1@stud.mw.tu-muenchen.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

$Revision: 1.1 $
$Date: 1999/05/02 07:59:46 $

}

unit Yaccmisc;

(* 4-22-98 RZ *)

interface

function LoadFromFile( var Dest; Count: Word; const FileName: string ): pointer;
procedure SaveToFile( var Source; Count: Word; const FileName: string);

type ResFileType = file of byte;

procedure WriteRCResHeader( var f: ResFileType; ResName: string; Size: word );
                                           {Size in bytes!}
    {first char of ResName shouldn't be Char($FF) }
procedure WriteRCResContent( var f: ResFileType; c : array of integer );

type  RCResFileType = TextFile;

procedure OpenRCStrFile( var f: RCResFileType; const Name: string );
procedure WriteRCResStrContent( var f: RCResFileType; Ident: Word; const Content: string );
procedure CloseRCStrFile( var f: RCResFileType );

implementation
uses WinTypes, SysUtils, Classes,
     {$ifdef VER80} YaccDlg16, {$else} {$ifdef WIN32} YaccDlg32, {$endif} {$endif}
     YaccBase;

function LoadFromFile( var Dest; Count: Word; const FileName: string ): pointer;
  {alloc mem with GetMem( R, Size ) before}
var thelp : TMemoryStream;
    f : file;
begin
  Result := nil;
  thelp := TMemoryStream.Create;
  try
    {get FileSize}
    AssignFile( f, FileName );
    Reset( f );
    thelp.SetSize( FileSize( f ) );
    CloseFile( f );
    {check size}
    if Count <> thelp.Size then
      raise EWriteError.Create('YaccMisc.LoadFromFile: Sizes don''t match!');
    {load file into TMemoryStream}
    thelp.LoadFromFile( FileName );
    {copy file into mem}
    move( thelp.Memory^ , Dest, thelp.Size );
  finally
    {Dispose TMemoryStream}
    thelp.SetSize( 0 );
    thelp.Free;
  end;
  Result := @Dest;
end;

procedure SaveToFile( var Source; Count: Word; const FileName: string);
var thelp : TMemoryStream;
begin
  thelp := TMemoryStream.Create;
  try
    {Size is in Count}
    thelp.SetSize( Count );
    {copy}
    move( Source, thelp.Memory^ , thelp.Size );
    thelp.SaveToFile( FileName );
  finally
    {Dispose TMemoryStream}
    thelp.SetSize( 0 );
    thelp.Free;
  end;
end;

procedure WriteRCResHeader( var f: ResFileType; ResName: string; Size: word );
var i : word;
    Content: byte;
begin
{ information from from wotsit
  European Site:  http://www.wotsit.demon.co.uk
  US Site:        http://wotsit.simsware.com
  File : res.zip}
  Content := $FF; write( f, Content );
  Content := 10; write( f, Content );
   { Type     variable    Resource type
                          RT_RCDATA       10  Application-defined data}
  Content := 0 ; write( f, Content );           {Name      variable    Resource name or identifier }
  { The type and name can be numeric or textual. If the first byte is $FF, the subsequent two
      bytes are the numeric value. Otherwise, the first byte is the first character of the
      ANSI string.}
  for i := 1 to length( ResName ) do
    begin
    Content := byte( UpCase( ResName[ i ] ) ) ;
    write( f, Content );
    end;
  Content := 0; write( f, Content ); {Endbyte of PChar}
  Content := $30; write( f, Content ); {Flags    2   Discardable=$1000, Moveable=$0010, Pure=$0020, Preload=$0040}
  Content := $0; write( f, Content );
  Content := lo(Size); write( f, Content );
  Content := hi(Size); write( f, Content );{Size  4  Size of the resource data that immediately follow the header}
  Content := 0; write( f, Content );
  Content := 0; write( f, Content );
end;

procedure WriteRCResContent( var f: ResFileType; c : array of integer );
type t = array[ 0 .. 65520 ] of byte;
var i : word;
    Content1, Content2 : byte;
begin
  for i := Low( c ) to High( c ) do
    begin
    Content1 := lo(c[ i ]);
    Content2 := hi(c[ i ]);
    write( f, Content1, Content2 );
    end
end;

var
  LastIdent : Longint;
  LastContent : String;

procedure OpenRCStrFile( var f: RCResFileType; const Name: string );
begin
  RenameIfExists( Name );
  AssignFile( f, Name );
  Rewrite( f );
  Writeln( f, '/**************************************************************************' );
  Writeln( f, '  ' + Name );
  Writeln( f, '    produced by DYacc' );
  Writeln( f, '    at ' + DateTimeToStr( Now ) );
  Writeln( f, '    from ' + yfilename );
  Writeln( f, '**************************************************************************/' );
  Writeln( f );
  Writeln( f, 'STRINGTABLE' );
  Writeln( f, '{' );
  LastIdent := -1;
  LastContent := '';
end;

function CutLastBackslash( const s: string ): string;
begin
  Result := copy( s, 1, 245 );
  if length( s ) > 245 then
    Result[ 245 ] := '';
  while Result[ length( Result ) ] = '\' do
    Result := copy( Result, 1, length( Result ) - 1 );
end;

function StripActions( const s: string ): string;
var i : byte;
begin
  Result := '';
  i := 1;
  while i <= length( s ) do
    begin
    if s[ i ] = '$' then
      begin
      while s[ i ] in [ '$', '0', '1' .. '9', ' ', ':' ] do
        begin
        i := i + 1;
        if i > length( s ) then
          exit;
        end;
      Result := Result + s[i];
      end
    else
      Result := Result + s[i];
    i := i + 1;
    end;
end;

const
  FmtStrLine : string[15] = ' %d, "%s"';

procedure WriteRCResStrContent( var f: RCResFileType; Ident: Word; const Content: string );
begin
  if LastIdent = Ident then
    LastContent := LastContent + '\n' + StripActions( Content )
  else
    begin
    if LastIdent <> -1 then
      writeln( f, Format( FmtStrLine, [ LastIdent + RuleStrBaseCount,
                                        CutLastBackslash( LastContent ) ] ) );
    LastContent := StripActions( Content );
    end;
  LastIdent := Ident;
end;

procedure CloseRCStrFile( var f: RCResFileType );
begin
  if LastIdent <> -1 then
    Writeln( f, Format( FmtStrLine, [ LastIdent + RuleStrBaseCount,
                                      CutLastBackslash( LastContent ) ] ) );

  Writeln( f, '}' );
  Writeln( f );
  Flush( f );
  CloseFile( f );
end;

initialization
end.
