{
  TP Yacc message and error handling module 2-5-91 AG
  Note: this module should be USEd by any module using the heap during
        initialization, since it installs a heap error handler (which
        terminates the program with fatal error `memory overflow').


  Copyright (c) 1990-92  Albert Graef <ag@muwiinfa.geschichte.uni-mainz.de>
  Copyright (C) 1998     Robert Zierer <rz1@stud.mw.tu-muenchen.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

$Revision: 1.1 $
$Date: 1999/05/02 07:59:46 $

}

unit YaccMsgs;

(* 2-5-91 AG *)
(* 4-22-98 RZ *)

interface

var errors, warnings : Integer;
  (* - current error and warning count *)
procedure error(msg : String);
  (* - print current input line and error message (pos denotes position to
       mark in source file line) *)
procedure warning(msg : String);
  (* - print warning message *)
procedure fatal(msg : String);
(* - writes a fatal error message, erases Yacc output file and terminates
     the program with errorlevel 1 *)
procedure beep;

procedure write_sign_on;
procedure write_usage;
procedure write_options;
procedure write_defaults;
procedure write_lstfileheader;

const

(* shortform sign-on and usage message: *)

version = 'TP Yacc Version 3.0ag.0 [May 98;$Revision: 1.1 $;$Date: 1999/05/02 07:59:46 $]';
sign_on = version+', Copyright (c) 1990-92 Albert Graef';
usage   = 'Usage: YACC [options] yacc-file[.Y] [output-file[.PAS]]';
options = 'Options: /v, /d, /te, /tr';

(* command line error messages: *)

invalid_option                  = 'invalid option ';
illegal_no_args                 = 'illegal number of parameters';

(* syntax errors: *)

open_comment_at_eof             = '101: open comment at end of file';
missing_string_terminator       = '102: missing string terminator';
rcurl_expected                  = '103: %} expected';
rbrace_expected                 = '104: } expected';
rangle_expected                 = '105: > expected';
ident_expected                  = '106: identifier expected';
error_in_def                    = '110: error in definition';
error_in_rule                   = '111: error in rule';
syntax_error 			= '112: syntax error';
unexpected_eof                  = '113: unexpected end of file';

(* semantic errors: *)

nonterm_expected                = '201: nonterminal expected';
literal_expected                = '202: literal expected';
double_tokennum_def             = '203: literal already defined';
unknown_identifier              = '204: unknown identifier';
type_error                      = '205: type error';
range_error                     = '206: range error';
empty_grammar 			= '207: empty grammar?';

(* fatal errors: *)

cannot_open_file 		= 'FATAL: cannot open file ';
write_error                     = 'FATAL: write error';
mem_overflow 			= 'FATAL: memory overflow';
intset_overflow 		= 'FATAL: integer set overflow';
sym_table_overflow 		= 'FATAL: symbol table overflow';
nt_table_overflow 		= 'FATAL: nonterminal table overflow';
lit_table_overflow 		= 'FATAL: literal table overflow';
type_table_overflow 		= 'FATAL: type table overflow';
prec_table_overflow 		= 'FATAL: precedence table overflow';
rule_table_overflow 		= 'FATAL: rule table overflow';
state_table_overflow 		= 'FATAL: state table overflow';
item_table_overflow 		= 'FATAL: item table overflow';
trans_table_overflow 		= 'FATAL: transition table overflow';
redn_table_overflow 		= 'FATAL: reduction table overflow';

internal_error                  = 'FATAL: internal error';
file_not_exist                  = 'FATAL: file doesn''t exist';
no_filename                     = 'FATAL: no filename specified';

implementation

uses SysUtils, Forms, YaccBase, YaccLib,
     {$ifdef VER80} YaccDlg16 {$else} {$ifdef WIN32} YaccDlg32 {$endif} {$endif}
     ;

procedure position(
            lineNo : integer;
            line : String;
            pos : integer);
  (* writes a position mark of the form
     lineno: line
               ^
     on f with the caret ^ positioned at pos in line
     a subsequent write starts at the next line, indented with tab *)
  var
    line1, line2 : String;
  begin
    (* this hack handles tab characters in line: *)
    line1 := intStr(lineNo)+': '+line;
    line2 := blankStr(intStr(lineNo)+': '+copy(line, 1, pos-1));
    GiveOutLine( omBoth, line1 );
    GiveOutLine( omBoth, line2 + '^' );
    if GenHtml then
      write( yyLst, '&nbsp;' );
    GiveOutChar( omBoth, tab );
  end(*position*);

procedure error(msg : String);
  begin
    inc(errors);
    GiveOutLine( omBoth, '' );
    position( lno, line, cno-tokleng );
    GiveOutLine( omBoth, msg );
  end(*error*);

procedure warning(msg : String);
  begin
    inc(warnings);
    GiveOutLine( omBoth, '' );
    position( lno, line, cno-tokleng );
    GiveOutLine( omBoth, msg );
  end(*warning*);

procedure fatal(msg : String);
  begin
    try
      GiveOutLine( omBoth, '' );
      GiveOutLine( omBoth, msg );
    finally
      {$I-}close(yyin); close(yyout); close(yycod); close(yylst); erase(yyout);{$I+}
      raise EYaccException.Create( msg );
    end;
  end(*fatal*);

procedure beep;
const i : word = 0;
begin
  if not yydebug then
    begin
    inc(i);
    if (i mod 5 ) = 0 then
      GiveOutChar( omScreen, ' ');
    GiveOutString( omScreen, '.');
    end;
  Application.ProcessMessages;
end;

procedure write_sign_on;
begin
  GiveOutLine( omScreen, version );
  GiveOutLine( omScreen, ' Copyright (c) 1990-92 Albert Graef');
  GiveOutLine( omScreen, ' Modified by Robert Zierer (c) 1996-98');
end;

procedure write_usage;
begin
  GiveOutLine( omScreen, usage );
  write_options;
end;

procedure write_options;
begin
  GiveOutLine( omScreen, 'Option  Description' );
  GiveOutLine( omScreen, '/V      Verbose mode; generates *.lst file' );
  GiveOutLine( omScreen, '/D      Debug mode; sets {$define yydebug} in source' );

  GiveOutLine( omScreen, '/TE     Output *.dat files with the tables rather than' );
  GiveOutLine( omScreen, '         store them in typed consts' );
  GiveOutLine( omScreen, '         sets {$define yyExternTables} in source' );
  GiveOutLine( omScreen, '/TR     Output files with the tables in a separate *.res' );
  GiveOutLine( omScreen, '         file; sets {$define yyResourceTables} in source' );
  GiveOutLine( omScreen, '/TRprefix  ditto, but prefixes res-names with prefix ' );

  GiveOutLine( omScreen, '/SS     Assumes a statik stack for yyv;' );
  GiveOutLine( omScreen, '         accesses yyv by ''yyv''' );
  GiveOutLine( omScreen, '/SD     Assumes a dynamic stack for yyv;' );
  GiveOutLine( omScreen, '         accesses yyv by ''yyv^''' );

  GiveOutLine( omScreen, '/RSR=d,fn Give out the RuleStrings in a *.RC resource-file' );
  GiveOutLine( omScreen, '           d is the startvalue for the string-ID' );
  GiveOutLine( omScreen, '           fn is the filename' );

  GiveOutLine( omScreen, '/RSP    Give out the RuleStrings into parser-file' );

  GiveOutLine( omScreen, '/RST=fn Give out the RuleStrings in a *.txt-file' );
  GiveOutLine( omScreen, '         fn is the filename' );
  GiveOutLine( omScreen, '   RuleStrings only in verbose mode' );

  GiveOutLine( omScreen, '/ND     For use in batch mode; waits not for confirmation' );
  GiveOutLine( omScreen, '         to close the dialog' );

  GiveOutLine( omScreen, '/HTML   Generates *.lst file as HTML-File' );

  GiveOutLine( omScreen, '/Ccodf  Uses codf instead of yyparse.cod file' );
end;

procedure write_defaults;
begin
  GiveOutLine( omScreen, 'Defaults are :' );
  GiveOutLine( omScreen, '  Non verbose, non debug, normal tables, statik stack' );
  GiveOutLine( omScreen, '  (/SS), no batch mode, no HTML Listing, no RuleStrings' );
end;

procedure write_lstfileheader;
begin
  if GenHtml then
    begin
    writeln( yyLst, '<HTML>');
    writeln( yyLst, '<HEAD>');
    writeln( yyLst, ' <META HTTP-EQUIV="Content-Type" CONTENT="text/html; charset=iso-8859-1">');
    writeln( yyLst, '<TITLE>DYacc-generated *.Lst-file for ', yfilename, '</TITLE>');
    writeln( yyLst, '</HEAD>');
    writeln( yyLst, '<BODY BGCOLOR="#FFFFFF">');
    end;
  if GenHtml then
    writeln( yyLst, '<CENTER>'+version+'</CENTER>' )
  else
    writeln( yyLst, version );
  if GenHtml then
    write( yylst, '<OL><UL>');
  GiveOutLine( omBoth, DateTimeToStr( Now ) );
  GiveOutLine( omBoth, 'Input : ' + yfilename );
  GiveOutLine( omBoth, 'Output : ' + pasfilename );
  GiveOutLine( omBoth, 'Code Template : ' + codfilename );
  GiveOutLine( omBoth, 'Current options :' );
  if GenHtml then
    writeln( yyLst, '<UL><OL>');
  if verbose then
    if GenHtml then
      GiveOutLine( omBoth, '  Generate HTML-Listing' )
    else
      GiveOutLine( omBoth, '  Generate normal Listing' )
  else
    GiveOutLine( omBoth, '  Generate no listing' );
  if debug then
    GiveOutLine( omBoth, '  Debug mode' )
  else
    GiveOutLine( omBoth, '  Non debug mode' );
  if BatchMode then
    GiveOutLine( omBoth, '  Batch mode' )
  else
    GiveOutLine( omBoth, '  Dialog mode' );
  case TableType of
    ttNormal :
      GiveOutLine( omBoth, '  TableType: Normal' );
    ttExtern :
      GiveOutLine( omBoth, '  TableType: Extern' );
    ttResource :
      begin
      GiveOutLine( omBoth, '  TableType: Resource' );
      GiveOutLine( omBoth, '    TableNamePrefix : '''+ResTablePrefix+'''');
      end;
  end;
  case StackType of
    stStatic :
      GiveOutLine( omBoth, '  Type of yyv: Static (means yyv[i])' );
    stDynamic :
      GiveOutLine( omBoth, '  Type of yyv: Dynamic (means yyv^[i])' );
  end;
  if RuleStrings then
    begin
    if verbose then
      begin
      GiveOutLine( omBoth, '  Give out RuleStrings' );
      case RuleStringPlace of
        rs_Parser:
          begin
          GiveOutLine( omBoth, '    into parserfile ''' + pasFilename + '''' );
          end;
        rs_Text:
          begin
          GiveOutLine( omBoth, '    to textfile ''' + RuleStrFilename + '''' );
          end;
        rs_RC:
          begin
          GiveOutLine( omBoth, '    to resourcefile ''' + RuleStrFilename + '''' );
          GiveOutLine( omBoth, '    starting value is ' + IntToStr( RuleStrBaseCount ) );
          end;
        else
          ;
      end;
      end
    else
      GiveOutLine( omBoth, '  no RuleStrings as not in verbose mode' );
    end;
  if GenHtml then
    writeln( yyLst, '</UL></OL></UL>' );
end;

{$F+}
function heapErrorHandler ( size : Word ) : Integer; far;
{$F-}
  begin
    heapErrorHandler := 0;
    if size>0 then
      fatal(mem_overflow) (* never returns *)
    else
      heapErrorHandler := 1
  end(*heapErrorHandler*);

begin
  errors := 0; warnings := 0;
  (* install heap error handler: *)
{$ifdef VER80}
  heapError := @heapErrorHandler;
{$endif}
end(*YaccMsgs*).

