{
  Some Array Types for the Yacc parser.


  Copyright (C) 1998     Robert Zierer <rz1@stud.mw.tu-muenchen.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

$Revision: 1.2 $
$Date: 1999/06/29 19:00:50 $

}

{$undef DeallocTables}

unit Yacctypes;

(* 4-22-98 RZ *)

interface
uses SysUtils, WinTypes, Dynarray;

procedure DeallocateTables;

type
  EDynLimArrayConcrete = class(Exception);

const
  Index_Error_write = 'Indexing error while writing to array !';
  Index_Error_read  = 'Indexing error while reading to array !';
  Index_Error_access = 'Indexing error while accessing to array !';

{=================================================================================}

type

  {a redimensionable array linking itself into a list}
  TDynLimArrayRes = class(TDynLimArray)
     private
       NextObject : TDynLimArrayRes;
     protected
     public
       constructor create;
       destructor destroy; override;
     end;

{=================================================================================}

  {a redimensionable array with elements of type Integer(Pascal) }
  TDynLimArray_Integer = class(TDynLimArrayRes)
     protected
       procedure SetElement( Index: longint; value: Integer );
       function GetElement( Index: longint ): Integer;
     public
       property Element[ Index: longint ]: Integer read GetElement write SetElement; default;
       constructor Create;
     end;

{=================================================================================}

  {a redimensionable array with elements of type Boolean(Pascal) }
  TDynLimArray_Boolean = class(TDynLimArrayRes)
     protected
       procedure SetElement( Index: longint; value: Boolean );
       function GetElement( Index: longint ): Boolean;
     public
       property Element[ Index: longint ]: Boolean read GetElement write SetElement; default;
       constructor Create;
     end;

{=================================================================================}

implementation
uses WinProcs, Forms, YaccTables, YaccBase, YaccDlg16, YaccMsgs;

{=================================================================================}

const Root : TDynLimArrayRes = nil;

{=================================================================================}

constructor TDynLimArrayRes.Create;
begin
  inherited Create;
  {make a list, so that the Objects can be deallocated on ProgramExit}
  NextObject := Root;
  Root := self;
end;

destructor TDynLimArrayRes.Destroy;
var p : TDynLimArrayRes;
begin
  if Root = self then begin
    Root := NextObject;
  end else begin
    p := Root;
    while (p <> nil) and (p.NextObject <> self) do
      p := p.NextObject;
    if p.NextObject = self then
      p.NextObject := NextObject;
  end;
  {NextObject := nil}
  inherited Destroy;
end;

{=================================================================================}

constructor TDynLimArray_Integer.Create;
begin
  inherited Create;
  ElementSize := System.Sizeof(Integer);
end;

procedure TDynLimArray_Integer.SetElement( Index: longint; value: Integer );
var temp : ^Integer;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_write);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    temp^ := value;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

function TDynLimArray_Integer.GetElement( Index: longint ): Integer;
var temp : ^Integer;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_read);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    result := temp^;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

{=================================================================================}

constructor TDynLimArray_Boolean.Create;
begin
  inherited Create;
  ElementSize := System.Sizeof(Boolean);
end;

procedure TDynLimArray_Boolean.SetElement( Index: longint; value: Boolean );
var temp : ^Boolean;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_write);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    temp^ := value;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;

function TDynLimArray_Boolean.GetElement( Index: longint ): Boolean;
var temp : ^Boolean;
begin
  if (Index > HighIndex) or (Index < LowIndex) then
    raise EDynLimArrayConcrete.Create(Index_Error_read);
  temp := GetElementPtr( Index );
  if temp <> nil then begin
    result := temp^;
  end else
    raise EDynLimArrayConcrete.Create(Index_Error_access);
end;


{=================================================================================}

procedure DeallocateTables;
var tmp, old : PropList;
    p : IntSetPtr;
var i, k : longint;
    ItemDeallocated : boolean;
begin
{$ifdef DeallocTables}
  {RZ: this should free all memory the prog allocates, on normal program termination;
     not tested what happens, when an Exception occurs (EFault,..) or an YaccMsgs.Error;
     (Note: seems that the VCL allocates 32 and 128 bytes of memory at
            addr $000E:0FF4 (unit classes ??) without deallocating it;
            supposing the fault in Classes.NewStrItem or an local routine before its impl)}
  GiveOutLine( omScreen, 'Deallocating Tables' );
  if Root = nil then exit;
  try
    try

      {DeAlloc PropTable items (PropList)}
        {lookaheadtable items (PropEntry.symset) are deallocated later}
      for i := 1 to n_items do
        if prop_table[i] <> nil then
          begin
          if (i mod 500) = 0 then YaccMsgs.Beep;
          {get next node}
          tmp := prop_table[i]^.next;
          {the root-node}
          dispose( prop_table[i] );
          {the leaves}
          while tmp <> nil do
            begin
            old := tmp;
            tmp := tmp^.next;
            dispose( old );
            end;
          end;

      {DeAlloc SymPropTable items (PString)}
      for i := sym_table.lowIndex to sym_table.highIndex do
        begin
        if (i mod 500) = 0 then YaccMsgs.Beep;
        if sym_table[i].pname <> nil then
          DisposeStr( sym_table[i].pname );
        end;

      {DeAlloc ClosureTable items (IntSetPtr)}
      for i := 1 to n_nts do
        begin
        if (i mod 500) = 0 then YaccMsgs.Beep;
        if closure_table[i] <> nil then
          FreeMem( closure_table[i], (max_elems+1)*sizeOf(Integer));
        end;

      {DeAlloc FirstSetTable items (IntSetPtr)}
      for i := 1 to n_nts do
        begin
        if (i mod 500) = 0 then YaccMsgs.Beep;
        if first_set_table[i] <> nil then
          FreeMem( first_set_table[i], (max_elems+1)*sizeOf(Integer));
        end;

      {DeAlloc LookaheadTable items (IntSetPtr)}
      for i := 1 to n_items do
        begin
        if (i mod 500) = 0 then YaccMsgs.Beep;
        if lookahead_table[i] <> nil then
          begin
          {make sure, FreeMem won't change the pointer as we need it later}
          p := lookahead_table[i];
          FreeMem( p, (max_elems+1)*sizeOf(Integer));
          end;
        end;

      {DeAlloc RednTable[].symset items (IntSetPtr);
       some of them are already deallocated via lookahead_table}
      for i := 1 to n_redns do
        begin
        if (i mod 40) = 0 then YaccMsgs.Beep;
        p := redn_table[i].symset;
        if p <> nil then
          begin
          {search through lookahead_table; if the same is in --> next i, if not --> del}
          ItemDeallocated := false;
          for k := 1 to n_items do
            if p = lookahead_table[k] then
              begin
              ItemDeallocated := true;
              break; {exit for-loop}
              end;
          if not ItemDeallocated then
            {not yet deallocated (or lookahead_table empty)}
            FreeMem( p, (max_elems+1)*sizeOf(Integer));
          {else already deallocated via lookahead_table}
          end;
        end;

      {DeAlloc RuleTable items (RuleRecPtr)}
      for i := 1 to n_rules do
        begin
        if (i mod 500) = 0 then YaccMsgs.Beep;
        if rule_table[i] <> nil then
          FreeMem( rule_table[i], 2*sizeOf(Integer)+rule_table[i]^.rhs_len*sizeOf(Integer) );
        end;

    finally
      {deallocate arrays;
       could have done that like in the initialization part of YaccTables, but why should I?}
      while Root <> nil do
        Root.Destroy;
    end;
  except
    on Exception do ;
  end;
{$endif}
end;

{*****************************************************************************}
{*****************************************************************************}
var OldExitProc: pointer;

{$F+}
procedure NewExitProc; far;
{$F-}
begin
  try
    if Root <> nil then
      DeallocateTables;
  finally
    ExitProc := OldExitProc;
  end;
end;

initialization
{$ifdef DeallocTables}
  OldExitProc := ExitProc;
  ExitProc := @NewExitProc;
{$endif}
end.
